@@ -4,24 +4,25 @@
 # processes which in turn handle the I/O and application process.
 # Listener sockets are started in the master process and shared with
 # forked worker children.
+#
+# Users do not need to know the internals of this class, but reading the
+# {source}[http://bogomips.org/unicorn.git/tree/lib/unicorn/http_server.rb]
+# is education for programmers wishing to learn how \Unicorn works.
+# See Unicorn::Configurator for information on how to configure \Unicorn.
 class Unicorn::HttpServer
+  # :stopdoc:
   attr_accessor :app, :request, :timeout, :worker_processes,
                 :before_fork, :after_fork, :before_exec,
                 :listener_opts, :preload_app,
                 :reexec_pid, :orig_app, :init_listeners,
                 :master_pid, :config, :ready_pipe, :user
   attr_reader :pid, :logger
-
-  # :stopdoc:
   include Unicorn::SocketHelper
   include Unicorn::HttpResponse
 
   # backwards compatibility with 1.x
   Worker = Unicorn::Worker
 
-  # prevents IO objects in here from being GC-ed
-  IO_PURGATORY = []
-
   # all bound listener sockets
   LISTENERS = []
 
@@ -73,19 +74,18 @@ class Unicorn::HttpServer
   #   Unicorn::HttpServer::START_CTX[0] = "/home/bofh/1.9.2/bin/unicorn"
   START_CTX = {
     :argv => ARGV.map { |arg| arg.dup },
-    :cwd => lambda {
-        # favor ENV['PWD'] since it is (usually) symlink aware for
-        # Capistrano and like systems
-        begin
-          a = File.stat(pwd = ENV['PWD'])
-          b = File.stat(Dir.pwd)
-          a.ino == b.ino && a.dev == b.dev ? pwd : Dir.pwd
-        rescue
-          Dir.pwd
-        end
-      }.call,
     0 => $0.dup,
   }
+  # We favor ENV['PWD'] since it is (usually) symlink aware for Capistrano
+  # and like systems
+  START_CTX[:cwd] = begin
+    a = File.stat(pwd = ENV['PWD'])
+    b = File.stat(Dir.pwd)
+    a.ino == b.ino && a.dev == b.dev ? pwd : Dir.pwd
+  rescue
+    Dir.pwd
+  end
+  # :stopdoc:
 
   # Creates a working server on host:port (strange things happen if
   # port isn't a Number).  Use HttpServer::run to start the server and
@@ -96,8 +96,8 @@ class Unicorn::HttpServer
     @request = Unicorn::HttpRequest.new
     self.reexec_pid = 0
     options = options.dup
-    self.ready_pipe = options.delete(:ready_pipe)
-    self.init_listeners = options[:listeners] ? options[:listeners].dup : []
+    @ready_pipe = options.delete(:ready_pipe)
+    @init_listeners = options[:listeners] ? options[:listeners].dup : []
     options[:use_defaults] = true
     self.config = Unicorn::Configurator.new(options)
     self.listener_opts = {}
@@ -118,35 +118,7 @@ class Unicorn::HttpServer
 
   # Runs the thing.  Returns self so you can run join on it
   def start
-    BasicSocket.do_not_reverse_lookup = true
-
-    # inherit sockets from parents, they need to be plain Socket objects
-    # before they become Kgio::UNIXServer or Kgio::TCPServer
-    inherited = ENV['UNICORN_FD'].to_s.split(/,/).map do |fd|
-      io = Socket.for_fd(fd.to_i)
-      set_server_sockopt(io, listener_opts[sock_name(io)])
-      IO_PURGATORY << io
-      logger.info "inherited addr=#{sock_name(io)} fd=#{fd}"
-      server_cast(io)
-    end
-
-    config_listeners = config[:listeners].dup
-    LISTENERS.replace(inherited)
-
-    # we start out with generic Socket objects that get cast to either
-    # Kgio::TCPServer or Kgio::UNIXServer objects; but since the Socket
-    # objects share the same OS-level file descriptor as the higher-level
-    # *Server objects; we need to prevent Socket objects from being
-    # garbage-collected
-    config_listeners -= listener_names
-    if config_listeners.empty? && LISTENERS.empty?
-      config_listeners << Unicorn::Const::DEFAULT_LISTEN
-      init_listeners << Unicorn::Const::DEFAULT_LISTEN
-      START_CTX[:argv] << "-l#{Unicorn::Const::DEFAULT_LISTEN}"
-    end
-    config_listeners.each { |addr| listen(addr) }
-    raise ArgumentError, "no listeners" if LISTENERS.empty?
-
+    inherit_listeners!
     # this pipe is used to wake us up from select(2) in #join when signals
     # are trapped.  See trap_deferred.
     init_self_pipe!
@@ -160,7 +132,7 @@ class Unicorn::HttpServer
 
     self.master_pid = $$
     build_app! if preload_app
-    maintain_worker_count
+    spawn_missing_workers
     self
   end
 
@@ -279,10 +251,9 @@ class Unicorn::HttpServer
 
     proc_name 'master'
     logger.info "master process ready" # test_exec.rb relies on this message
-    if ready_pipe
-      ready_pipe.syswrite($$.to_s)
-      ready_pipe.close rescue nil
-      self.ready_pipe = nil
+    if @ready_pipe
+      @ready_pipe.syswrite($$.to_s)
+      @ready_pipe = @ready_pipe.close rescue nil
     end
     begin
       reap_all_workers
@@ -333,10 +304,8 @@ class Unicorn::HttpServer
           reexec
         end
       end
-    rescue Errno::EINTR
     rescue => e
-      logger.error "Unhandled master loop exception #{e.inspect}."
-      logger.error e.backtrace.join("\n")
+      Unicorn.log_error(@logger, "master loop error", e)
     end while true
     stop # gracefully shutdown all workers on our way out
     logger.info "master complete"
@@ -364,6 +333,22 @@ class Unicorn::HttpServer
                                 Unicorn::TeeInput : Unicorn::StreamInput
   end
 
+  def client_body_buffer_size
+    Unicorn::TeeInput.client_body_buffer_size
+  end
+
+  def client_body_buffer_size=(bytes)
+    Unicorn::TeeInput.client_body_buffer_size = bytes
+  end
+
+  def trust_x_forwarded
+    Unicorn::HttpParser.trust_x_forwarded?
+  end
+
+  def trust_x_forwarded=(bool)
+    Unicorn::HttpParser.trust_x_forwarded = bool
+  end
+
   private
 
   # wait for a signal hander to wake us up and then consume the pipe
@@ -387,7 +372,7 @@ class Unicorn::HttpServer
         self.pid = pid.chomp('.oldbin') if pid
         proc_name 'master'
       else
-        worker = WORKERS.delete(wpid) and worker.tmp.close rescue nil
+        worker = WORKERS.delete(wpid) and worker.close rescue nil
         m = "reaped #{status.inspect} worker=#{worker.nr rescue 'unknown'}"
         status.success? ? logger.info(m) : logger.error(m)
       end
@@ -410,7 +395,6 @@ class Unicorn::HttpServer
 
     if pid
       old_pid = "#{pid}.oldbin"
-      prev_pid = pid.dup
       begin
         self.pid = old_pid  # clear the path for a new pid file
       rescue ArgumentError
@@ -445,22 +429,17 @@ class Unicorn::HttpServer
     proc_name 'master (old)'
   end
 
-  # forcibly terminate all workers that haven't checked in in timeout
-  # seconds.  The timeout is implemented using an unlinked File
-  # shared between the parent process and each worker.  The worker
-  # runs File#chmod to modify the ctime of the File.  If the ctime
-  # is stale for >timeout seconds, then we'll kill the corresponding
-  # worker.
+  # forcibly terminate all workers that haven't checked in in timeout seconds.  The timeout is implemented using an unlinked File
   def murder_lazy_workers
     t = @timeout
     next_sleep = 1
+    now = Time.now.to_i
     WORKERS.dup.each_pair do |wpid, worker|
-      stat = worker.tmp.stat
-      # skip workers that disable fchmod or have never fchmod-ed
-      stat.mode == 0100600 and next
-      diff = Time.now - stat.ctime
-      if diff <= t
-        tmp = t - diff
+      tick = worker.tick
+      0 == tick and next # skip workers that are sleeping
+      diff = now - tick
+      tmp = t - diff
+      if tmp >= 0
         next_sleep < tmp and next_sleep = tmp
         next
       end
@@ -471,17 +450,35 @@ class Unicorn::HttpServer
     next_sleep
   end
 
+  def after_fork_internal
+    @ready_pipe.close if @ready_pipe
+    Unicorn::Configurator::RACKUP.clear
+    @ready_pipe = @init_listeners = @before_exec = @before_fork = nil
+
+    srand # http://redmine.ruby-lang.org/issues/4338
+
+    # The OpenSSL PRNG is seeded with only the pid, and apps with frequently
+    # dying workers can recycle pids
+    OpenSSL::Random.seed(rand.to_s) if defined?(OpenSSL::Random)
+  end
+
   def spawn_missing_workers
-    (0...worker_processes).each do |worker_nr|
-      WORKERS.values.include?(worker_nr) and next
-      worker = Worker.new(worker_nr, Unicorn::TmpIO.new)
+    worker_nr = -1
+    until (worker_nr += 1) == @worker_processes
+      WORKERS.value?(worker_nr) and next
+      worker = Worker.new(worker_nr)
       before_fork.call(self, worker)
-      WORKERS[fork {
-        ready_pipe.close if ready_pipe
-        self.ready_pipe = nil
+      if pid = fork
+        WORKERS[pid] = worker
+      else
+        after_fork_internal
         worker_loop(worker)
-      }] = worker
+        exit
+      end
     end
+    rescue => e
+      @logger.error(e) rescue nil
+      exit!
   end
 
   def maintain_worker_count
@@ -500,32 +497,34 @@ class Unicorn::HttpServer
     msg = case e
     when EOFError,Errno::ECONNRESET,Errno::EPIPE,Errno::EINVAL,Errno::EBADF
       Unicorn::Const::ERROR_500_RESPONSE
+    when Unicorn::RequestURITooLongError
+      Unicorn::Const::ERROR_414_RESPONSE
+    when Unicorn::RequestEntityTooLargeError
+      Unicorn::Const::ERROR_413_RESPONSE
     when Unicorn::HttpParserError # try to tell the client they're bad
       Unicorn::Const::ERROR_400_RESPONSE
     else
-      logger.error "Read error: #{e.inspect}"
-      logger.error e.backtrace.join("\n")
+      Unicorn.log_error(@logger, "app error", e)
       Unicorn::Const::ERROR_500_RESPONSE
     end
     client.kgio_trywrite(msg)
     client.close
     rescue
-      nil
   end
 
   # once a client is accepted, it is processed in its entirety here
   # in 3 easy steps: read request, call app, write app response
   def process_client(client)
-    r = @app.call(env = @request.read(client))
+    status, headers, body = @app.call(env = @request.read(client))
 
-    if 100 == r[0].to_i
+    if 100 == status.to_i
       client.write(Unicorn::Const::EXPECT_100_RESPONSE)
       env.delete(Unicorn::Const::HTTP_EXPECT)
-      r = @app.call(env)
+      status, headers, body = @app.call(env)
     end
-    # r may be frozen or const, so don't modify it
-    @request.headers? or r = [ r[0], nil, r[2] ]
-    http_response_write(client, r)
+    @request.headers? or headers = nil
+    http_response_write(client, status, headers, body)
+    client.close # flush and uncork socket immediately, no keepalive
   rescue => e
     handle_error(client, e)
   end
@@ -541,13 +540,12 @@ class Unicorn::HttpServer
     proc_name "worker[#{worker.nr}]"
     START_CTX.clear
     init_self_pipe!
-    WORKERS.values.each { |other| other.tmp.close rescue nil }
     WORKERS.clear
     LISTENERS.each { |sock| sock.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC) }
-    worker.tmp.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC)
     after_fork.call(self, worker) # can drop perms
     worker.user(*user) if user.kind_of?(Array) && ! worker.switched
     self.timeout /= 2.0 # halve it for select()
+    @config = nil
     build_app! unless preload_app
   end
 
@@ -556,6 +554,9 @@ class Unicorn::HttpServer
     Unicorn::Util.reopen_logs
     logger.info "worker=#{worker_nr} done reopening logs"
     init_self_pipe!
+    rescue => e
+      logger.error(e) rescue nil
+      exit!(77) # EX_NOPERM in sysexits.h
   end
 
   # runs inside each forked worker, this sits around and waits
@@ -565,35 +566,25 @@ class Unicorn::HttpServer
     ppid = master_pid
     init_worker_process(worker)
     nr = 0 # this becomes negative if we need to reopen logs
-    alive = worker.tmp # tmp is our lifeline to the master process
-    ready = LISTENERS
+    l = LISTENERS.dup
+    ready = l.dup
 
     # closing anything we IO.select on will raise EBADF
     trap(:USR1) { nr = -65536; SELF_PIPE[0].close rescue nil }
-    trap(:QUIT) { alive = nil; LISTENERS.each { |s| s.close rescue nil } }
+    trap(:QUIT) { worker = nil; LISTENERS.each { |s| s.close rescue nil }.clear }
     [:TERM, :INT].each { |sig| trap(sig) { exit!(0) } } # instant shutdown
     logger.info "worker=#{worker.nr} ready"
-    m = 0
 
     begin
       nr < 0 and reopen_worker_logs(worker.nr)
       nr = 0
 
-      # we're a goner in timeout seconds anyways if alive.chmod
-      # breaks, so don't trap the exception.  Using fchmod() since
-      # futimes() is not available in base Ruby and I very strongly
-      # prefer temporary files to be unlinked for security,
-      # performance and reliability reasons, so utime is out.  No-op
-      # changes with chmod doesn't update ctime on all filesystems; so
-      # we change our counter each and every time (after process_client
-      # and before IO.select).
-      alive.chmod(m = 0 == m ? 1 : 0)
-
-      ready.each do |sock|
+      worker.tick = Time.now.to_i
+      while sock = ready.shift
         if client = sock.kgio_tryaccept
           process_client(client)
           nr += 1
-          alive.chmod(m = 0 == m ? 1 : 0)
+          worker.tick = Time.now.to_i
         end
         break if nr < 0
       end
@@ -602,23 +593,21 @@ class Unicorn::HttpServer
       # we're probably reasonably busy, so avoid calling select()
       # and do a speculative non-blocking accept() on ready listeners
       # before we sleep again in select().
-      redo unless nr == 0 # (nr < 0) => reopen logs
+      unless nr == 0 # (nr < 0) => reopen logs (unlikely)
+        ready = l.dup
+        redo
+      end
 
       ppid == Process.ppid or return
-      alive.chmod(m = 0 == m ? 1 : 0)
 
       # timeout used so we can detect parent death:
-      ret = IO.select(LISTENERS, nil, SELF_PIPE, timeout) and ready = ret[0]
-    rescue Errno::EINTR
-      ready = LISTENERS
+      worker.tick = Time.now.to_i
+      ret = IO.select(l, nil, SELF_PIPE, @timeout) and ready = ret[0]
     rescue Errno::EBADF
       nr < 0 or return
     rescue => e
-      if alive
-        logger.error "Unhandled listen loop exception #{e.inspect}."
-        logger.error e.backtrace.join("\n")
-      end
-    end while alive
+      Unicorn.log_error(@logger, "listen loop error", e) if worker
+    end while worker
   end
 
   # delivers a signal to a worker and fails gracefully if the worker
@@ -626,7 +615,7 @@ class Unicorn::HttpServer
   def kill_worker(signal, wpid)
     Process.kill(signal, wpid)
     rescue Errno::ESRCH
-      worker = WORKERS.delete(wpid) and worker.tmp.close rescue nil
+      worker = WORKERS.delete(wpid) and worker.close rescue nil
   end
 
   # delivers a signal to each worker
@@ -656,7 +645,7 @@ class Unicorn::HttpServer
   def load_config!
     loaded_app = app
     logger.info "reloading config_file=#{config.config_file}"
-    config[:listeners].replace(init_listeners)
+    config[:listeners].replace(@init_listeners)
     config.reload
     config.commit!(self)
     kill_each_worker(:QUIT)
@@ -665,8 +654,8 @@ class Unicorn::HttpServer
     build_app! if preload_app
     logger.info "done reloading config_file=#{config.config_file}"
   rescue StandardError, LoadError, SyntaxError => e
-    logger.error "error reloading config_file=#{config.config_file}: " \
-                 "#{e.class} #{e.message} #{e.backtrace}"
+    Unicorn.log_error(@logger,
+        "error reloading config_file=#{config.config_file}", e)
     self.app = loaded_app
   end
 
@@ -700,5 +689,33 @@ class Unicorn::HttpServer
     SELF_PIPE.replace(Kgio::Pipe.new)
     SELF_PIPE.each { |io| io.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC) }
   end
-end
 
+  def inherit_listeners!
+    # inherit sockets from parents, they need to be plain Socket objects
+    # before they become Kgio::UNIXServer or Kgio::TCPServer
+    inherited = ENV['UNICORN_FD'].to_s.split(/,/).map do |fd|
+      io = Socket.for_fd(fd.to_i)
+      set_server_sockopt(io, listener_opts[sock_name(io)])
+      IO_PURGATORY << io
+      logger.info "inherited addr=#{sock_name(io)} fd=#{fd}"
+      server_cast(io)
+    end
+
+    config_listeners = config[:listeners].dup
+    LISTENERS.replace(inherited)
+
+    # we start out with generic Socket objects that get cast to either
+    # Kgio::TCPServer or Kgio::UNIXServer objects; but since the Socket
+    # objects share the same OS-level file descriptor as the higher-level
+    # *Server objects; we need to prevent Socket objects from being
+    # garbage-collected
+    config_listeners -= listener_names
+    if config_listeners.empty? && LISTENERS.empty?
+      config_listeners << Unicorn::Const::DEFAULT_LISTEN
+      @init_listeners << Unicorn::Const::DEFAULT_LISTEN
+      START_CTX[:argv] << "-l#{Unicorn::Const::DEFAULT_LISTEN}"
+    end
+    config_listeners.each { |addr| listen(addr) }
+    raise ArgumentError, "no listeners" if LISTENERS.empty?
+  end
+end
